#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

import sys, errno, os, stat, signal, json, socket, threading, struct, six
from hypervisor import *

import logging
import logging.handlers

logger = logging.getLogger(defines.VM_LOGGER_NAME)
logger.setLevel(logging.DEBUG)

handler = logging.handlers.SysLogHandler(address='/dev/log')
handler.setLevel(logging.INFO)

formatter = logging.Formatter('[%(levelname)s] synoabk_vm::%(filename)s:%(lineno)d(%(process)d,%(thread)d) %(message)s')
handler.setFormatter(formatter)

logger.addHandler(handler)

# run daemon
if __name__ == '__main__':
    if six.PY3:
        from importlib import reload
        reload(sys)
    elif six.PY2:
        reload(sys)
        sys.setdefaultencoding('utf-8')

    serverLock = threading.Lock()
    serverPool = {}

    def GetServer(inventory):
        serverLock.acquire()
        try:
            inventoryId = inventory['inventory_id']
            hosttype    = inventory['host_type']
            hostaddr    = inventory['host_addr']
            username    = inventory['auth_user']
            password    = inventory['auth_password']
            thumbprint  = inventory['auth_thumbprint']
            port        = inventory['port_webapi']
            protocol    = inventory['protocol']

            if inventoryId != 0:
                oldInventory, server, taskPool = serverPool.get(inventoryId) or (None, None, {})
            else:
                oldInventory, server, taskPool = None, None, {}

            if  hosttype == defines.HostType.ESXi.value or \
                hosttype == defines.HostType.vCenter.value:
                if  server == None or \
                    (oldInventory['host_addr'] != hostaddr or \
                    oldInventory['auth_user'] != username or \
                    oldInventory['auth_password'] != password or \
                    oldInventory['auth_thumbprint'] != thumbprint or \
                    oldInventory['protocol'] != protocol or \
                    oldInventory['port_webapi'] != port):
                    server = vsphere_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool, None, 0)
            elif hosttype == defines.HostType.HyperV.value or \
                    hosttype == defines.HostType.FailoverCluster.value:
                server = hyperv_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool, hosttype, None, 0)
            elif hosttype == defines.HostType.SCVMM.value:
                server = scvmm_server.Main(hostaddr, username, password, thumbprint, port, protocol, taskPool)
            else:
                raise Exception('Unknown Host Type %d' % hosttype)

            serverPool[inventoryId] = (inventory, server, taskPool)
        finally:
            serverLock.release()

        return server

    def HandleRequest(request):
        command = request['command']
        argument = request['argument']
        inventory = request['inventory']
        server = GetServer(inventory)
        logger.info(command)
        logger.info({k:v for k,v in argument.items() if k not in ['vm_user', 'vm_pwd']} if argument else None)
        return server.dispatch(command, argument)

    def HandleConnection(connection, address):
        try:
            while True:
                data = connection.recv(4)
                if len(data) != 4: break

                n = sum([(ord(data[i]) if type(data[i]) == str else data[i])*(256**i) for i in range(4)])

                data = connection.recv(n)
                if len(data) != n: break

                data = six.ensure_binary(json.dumps(HandleRequest(json.loads(six.ensure_str(data, 'utf-8', 'ignore')))))
                n = len(data)
                data = b''.join([struct.pack('B', int(n/(256**i))%256) for i in range(4)]) + data

                connection.sendall(data)
        finally:
            connection.close()

    def HandleSignal(sig, frame):
        if sig == signal.SIGPIPE:
            return
        serverPool.clear()
        os._exit(0)

    logger.info('Hypervisor Server Daemon Start! ^.^')

    utils.beUser(int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))

    try:
        os.unlink(defines.HYPERVISOR_SERVER_SOCKET)
    except Exception:
        pass

    try:
        os.makedirs(os.path.dirname(defines.HYPERVISOR_SERVER_SOCKET))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
    sock.bind(defines.HYPERVISOR_SERVER_SOCKET)
    sock.listen(16)

    os.chown(defines.HYPERVISOR_SERVER_SOCKET, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
    os.chmod(defines.HYPERVISOR_SERVER_SOCKET, stat.S_IRUSR + stat.S_IWUSR)

    try:
        os.mkdir(defines.VM_SYNOABK_TEMP)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise

    os.chown(defines.VM_SYNOABK_TEMP, int(os.environ['PKG_UID']), int(os.environ['PKG_GID']))
    os.chmod(defines.VM_SYNOABK_TEMP, stat.S_IRWXU + stat.S_IRGRP + stat.S_IXGRP + stat.S_IROTH + stat.S_IXOTH)

    for sig in (signal.SIGINT, signal.SIGTERM, signal.SIGQUIT, signal.SIGPIPE):
        signal.signal(sig, HandleSignal)

    while True:
        threading.Thread(target=HandleConnection, args=sock.accept()).start()
