# -*- coding: utf-8 -*-

import util
from render import render_size, render_dedup_rate
import re

def normallize_series(config):
	return {
		'label':              config[0],
		'backup_type':        config[1],
		'color':              config[2],
		'display_short_text': config[3],
		'display_text':       config[4],
	}

def is_dsm6():
	try:
		with open('/etc.defaults/VERSION', 'r') as f:
			s = f.read()
			return '6' == dict(re.findall(r'(\S+)=(?:"(.*?)"|[\S^"]+)', s)).get("majorversion", '')
	except:
		return False

def is_support_dsm_backup():
	try:
		with open('/etc.defaults/VERSION', 'r') as f:
			s = f.read()
			major = int(dict(re.findall(r'(\S+)=(?:"(.*?)"|[\S^"]+)', s)).get("majorversion", ''))
			minor = int(dict(re.findall(r'(\S+)=(?:"(.*?)"|[\S^"]+)', s)).get("minorversion", ''))
			return ((major == 7 and minor >= 1) or major > 7)
	except:
		return False

if is_dsm6():
	config_map = {
		config[0]: normallize_series(config)
		for config in [
			('',   0, '#3876BE', 'report:write_size', ''),
			('PC', 2, '#3397BB', 'common:device_pcs', 'common:device_personal_computer'),
			('PS', 3, '#F2AF2A', 'common:device_servers', 'common:device_physical_server'),
			('FS', 4, '#7ED321', 'common:device_fs', 'common:device_file_server'),
			('VM', 1, '#3FCEE6', 'common:device_vms', 'common:device_virtual_machine'),
		]
	}
else:
	configs = [
		('',   0, '#3876BE', 'report:write_size', ''),
		('PC', 2, '#2693FF', 'common:device_pcs', 'common:device_personal_computer'),
		('PS', 3, '#FAAF00', 'common:device_servers', 'common:device_physical_server'),
		('FS', 4, '#996DF2', 'common:device_fs', 'common:device_file_server'),
		('VM', 1, '#26BFBF', 'common:device_vms', 'common:device_virtual_machine'),
	]
	if is_support_dsm_backup():
		configs.append(('DS', 5, '#EB5E81', 'common:device_dsm', 'common:device_dsm_server'))

	config_map = {
		config[0]: normallize_series(config)
		for config in configs
	}

backup_types = ['PC', 'PS', 'VM']
if is_support_dsm_backup():
	backup_types.append('DS')


def collect_distinct_task_type(storage):
	using_device_type = set(task['task']['backup_type'] for task in storage['task_list'] if task['task']['backup_type'] in usage_valid_type)

	return [config_map[backup_type] for backup_type in backup_types
		if config_map[backup_type]['backup_type'] in using_device_type and get_storage_image_size(storage['task_list'], config_map[backup_type]['backup_type']) > 0]

def get_volume_disk_usage(volume):
	return sum([storage['total_size'] for storage in volume['storage_list']])

def find_storage_trend(report_data, vol_id):
	for storage_trend in report_data['storage_trend_data']:
		if storage_trend['vol_id'] == vol_id:
			return storage_trend['usage_list']
	return []

def get_total_write_size(report_data, vol_id):
	return sum([trend['write_size'] for trend in find_storage_trend(report_data, vol_id)])

usage_valid_type = set(config['backup_type'] for config in config_map.values() if config['label'] != 'FS')
def is_valid_storage(storage):
	for task in storage['task_list']:
		if task['usage'] <= 0:
			continue
		if task['task']['backup_type'] not in usage_valid_type:
			continue
		return True

	return storage['dedup_size'] == -1

def get_storage_image_size(task_list, backup_type=-1):
	if backup_type == -1:
		valid_type = usage_valid_type
	else:
		valid_type = set([backup_type])
	return sum([task['usage'] for task in task_list if task['usage'] > 0 and task['task']['backup_type'] in valid_type])

def break_storage_name(text):
	if util.output_format == 'pdf':
		# font size, width defined in css
		# lines = break_text(text, scaling(16.0) * 1.25, scaling(248), max_line=3)
		# return u'<br>'.join(lines)
		# NOTE: breakline handled by reportlab with -pdf-word-wrap: CJK
		return text
	else:
		return text

def generate_storage_data(storage):
	columns = [dict(values=[get_storage_image_size(storage['task_list'], config_map[key]['backup_type'])], **config_map[key])
		for key in backup_types if get_storage_image_size(storage['task_list'], config_map[key]['backup_type']) > 0
	]
	if storage['dedup_size'] == -1:
		return {
			'storage_id': storage['storage_id'],
			'share_name': break_storage_name(storage['share_name']),
			'is_compressed': False,
			'is_encrypted': False,
			'dedup': '',
			'compress': '',
			'dedup_size': '',
			'columns': [],
			# for html only
			'chart_id': u'storage-usage-chart-{}'.format(storage['storage_id']),
			'device_type_list': []
		}
	elif columns:
		transfer_size = get_storage_image_size(storage['task_list'])
		dedup_size = storage['dedup_size'] + storage.get('compressed_size', 0)
		return {
			'storage_id': storage['storage_id'],
			'share_name': break_storage_name(storage['share_name']),
			'is_compressed': storage.get('is_compressed', False),
			'is_encrypted': storage.get('is_encrypted', False),
			'dedup': render_dedup_rate(transfer_size, dedup_size),
			'compress': render_dedup_rate(dedup_size, storage['dedup_size']),
			'dedup_size': render_size(storage['dedup_size']),
			'columns': columns,
			# for html only
			'chart_id': u'storage-usage-chart-{}'.format(storage['storage_id']),
			'device_type_list': collect_distinct_task_type(storage),
		}

