#!/bin/python
# -*- coding: utf-8 -*-
import argparse
import csv
from datetime import datetime
import json
import os
import sys
import util
from util import StringManager
from hashlib import md5
from codecs import BOM_UTF8

def to_str(input):
	'''
	(bytes == str) in python2
	(bytes != str) in python3
	'''
	if str == type(input):
		return str(input)
	elif bytes == type(input):
		# for python3
		return str(input, 'utf-8')
	else:
		return str(input)

def get_datetime(timestamp):
	datetime_obj = datetime.fromtimestamp(timestamp)
	return datetime_obj.strftime('%Y-%m-%d %H:%M:%S')

def process_log_input_data(input_data):
	column_names = [to_str(StringManager.get('log:log_type').encode('utf-8')),
					to_str(StringManager.get('log:log_time').encode('utf-8')),
					to_str(StringManager.get('log:log_description').encode('utf-8'))]
	records = []
	template_log_string_cache = {}
	log_string_cache = {}
	log_level_string_cache = {}

	for log in input_data["data"]["logs"]:
		log_time = get_datetime(log.get('log_time', 0))
		other_params = log.get('other_params', {})
		log_type = log.get('log_type', 0)
		log_level = log.get('log_level', 0)

		if log_level_string_cache.get(log_level) is None:
			log_level_string_cache[log_level] = to_str(StringManager.get_log_type_string(log_level).encode('utf-8'))

		dict_string = (to_str(json.dumps(other_params)) + to_str(log_type)).encode('utf-8')
		dict_md5 = md5(dict_string).hexdigest()

		if log_string_cache.get(dict_md5) is None:
			if template_log_string_cache.get(log_type) is None:
				template_log_string_cache[log_type] = StringManager.get_log_template_string(log_type)

			template = template_log_string_cache[log_type]
			if 'timestamp_list' in other_params:
				str_timestamp_list = ', '.join([get_datetime(t) for t in other_params['timestamp_list']])
				template_rep = template.replace("{timestamp_list}", str_timestamp_list)
				del other_params['timestamp_list']
			else:
				template_rep = template

			log_string_cache[dict_md5] = to_str(StringManager.get_log_string_with_template(log_type, other_params, True, template_rep).encode('utf-8'))

		records.append([
			log_level_string_cache[log_level],
			log_time,
			log_string_cache[dict_md5],
		])

	return records, column_names

def get_result_task_source(task_config):
	if not task_config:
		return ""

	if task_config.get('device_list') is not None:
		name_arr = map(lambda device: device['host_name'], task_config["device_list"])
		return ",".join(str(x) for x in name_arr)
	elif task_config.get('share_name') is not None:
		return task_config["share_name"]
	elif task_config.get('device_name') is not None:
		return task_config["device_name"]
	else:
		return ""

def process_detail_input_data(input_data):
	column_names = [to_str(StringManager.get('log:task_log_type').encode('utf-8')),
					to_str(StringManager.get('log:task_log_description').encode('utf-8')),
					to_str(StringManager.get('log:task_log_time').encode('utf-8'))]
	records = []
	template_log_string_cache = {}
	log_string_cache = {}
	log_level_string_cache = {}

	for detail in input_data["data"]["result_detail_list"]:
		log_time = get_datetime(detail.get('log_time', 0))
		other_params = detail.get('other_params', {})
		log_type = detail.get('log_type', 0)
		log_level = detail.get('log_level', 0)

		if log_level_string_cache.get(log_level) is None:
			log_level_string_cache[log_level] = to_str(StringManager.get_log_type_string(log_level).encode('utf-8'))

		dict_string = (to_str(json.dumps(other_params)) + to_str(log_type)).encode('utf-8')
		dict_md5 = md5(dict_string).hexdigest()

		if log_string_cache.get(dict_md5) is None:
			if template_log_string_cache.get(log_type) is None:
				template_log_string_cache[log_type] = StringManager.get_log_template_string(log_type)

			template = template_log_string_cache[log_type]
			if 'timestamp_list' in other_params:
				str_timestamp_list = ', '.join([get_datetime(t) for t in other_params['timestamp_list']])
				template_rep = template.replace("{timestamp_list}", str_timestamp_list)
				del other_params['timestamp_list']
			else:
				template_rep = template

			log_string_cache[dict_md5] = to_str(StringManager.get_log_string_with_template(log_type, other_params, True, template_rep).encode('utf-8'))

		records.append([
			log_level_string_cache[log_level],
			log_string_cache[dict_md5],
			log_time,
		])

	return records, column_names

def process_result_input_data(input_data):
	column_names = [to_str(StringManager.get('component:task_status').encode('utf-8')),
					to_str(StringManager.get('task_info:task_type').encode('utf-8')),
					to_str(StringManager.get('task_info:task_name').encode('utf-8')),
					to_str(StringManager.get('log:task_start_time').encode('utf-8')),
					to_str(StringManager.get('log:task_finish_time').encode('utf-8')),
					to_str(StringManager.get('component:task_source').encode('utf-8'))]
	records = []
	status_string_cache = {}
	job_action_string_cache = {}

	for result in input_data["data"]["results"]:
		task_status = result.get('status', 0)
		task_type = result.get('job_action', 0)
		task_name = result.get('task_name', '')
		task_start_time = get_datetime(result.get('time_start', 0))
		task_finish_time = get_datetime(result.get('time_end', 0))
		task_source = get_result_task_source(result.get('task_config', {}))

		if status_string_cache.get(task_status) is None:
			status_string_cache[task_status] = StringManager.get_result_status_string(task_status)

		if job_action_string_cache.get(task_type) is None:
			job_action_string_cache[task_type] = StringManager.get_job_action_string(task_type)

		records.append([
			to_str(status_string_cache[task_status].encode('utf-8')),
			to_str(job_action_string_cache[task_type].encode('utf-8')),
			to_str(task_name.encode('utf-8')),
			task_start_time,
			task_finish_time,
			to_str(task_source.encode('utf-8')),
		])

	return records, column_names


def process_input_data(src_file):
	with open(src_file) as f:
		input_data = json.load(f)

		if input_data["data"].get("logs") is not None:
			return process_log_input_data(input_data)
		elif input_data["data"].get("results") is not None:
			return process_result_input_data(input_data)
		elif input_data["data"].get("result_detail_list") is not None:
			return process_detail_input_data(input_data)

	return [], []


def save_as_csv(records, column_names, dest_file):
	with open(dest_file, 'wb') as f:
		f.write(BOM_UTF8)

	with open(dest_file, 'a') as f:
		writer = csv.writer(f)
		writer.writerow(column_names)
		writer.writerows(records)


def parse_args():
	parser = argparse.ArgumentParser(description='Export log from JSON file')
	parser.add_argument('--src', dest='src_file', type=str, help='Absolute path of input files. "/a/b/c,/a/b/d"', required=True)
	parser.add_argument('--dest', dest='dest_file', type=str, help='Absolute path of output file', required=True)
	parser.add_argument('--resource-dir', default=os.getcwd(), help='Path contains string file, style, js script')
	parser.add_argument('--lang', default='enu', help='Language used in log.')

	args = parser.parse_args()
	return args


if __name__ == '__main__':
	args = parse_args()

	src_files = args.src_file
	dest_file = args.dest_file
	script_dir = os.path.dirname(os.path.abspath(__file__))
	resource_dir = args.resource_dir
	str_file = os.path.join(resource_dir, 'lang.js')
	js_file = os.path.join(script_dir, 'log.js')
	lang = args.lang
	all_records = []

	StringManager.initialize(lang, str_file, js_file)

	src_file_arr = src_files.split(",")
	src_ext = os.path.splitext(src_file_arr[0])[1]
	dest_ext = os.path.splitext(dest_file)[1]

	if src_ext != '.json':
		print('Error: wrong input format, need to be .json file')
		sys.exit(1)

	if dest_ext != '.csv':
		print('Error: wrong outputput format, need to be .csv file')
		sys.exit(1)

	for single_src_file in src_file_arr:
		records, column_names = process_input_data(single_src_file)
		all_records += records

	if all_records != [] and dest_ext == '.csv':
		save_as_csv(all_records, column_names, dest_file)
	else:
		print('Error: no records found.')

