import json
from datetime import datetime, timedelta
from functools import reduce

import util
from util import StringManager, generate_time_series, get_unit, escape

def scaling(n):
	if util.output_format == 'pdf':
		return n * 0.65
	else:
		return n

def v_scaling(n):
	if util.output_format == 'pdf':
		return n * 0.564
	else:
		return n

def render_size(size):
	unit_str, unit_value = get_unit(size)
	return '%.1f %s' % (size / unit_value, unit_str)

def render_dedup_rate(original_size, dedup_size):
		if dedup_size <= 0 or original_size <= 0:
			return '--'

		return '%.2fx' % (float(original_size) / dedup_size,)

def render_time_series(time_start, time_end, time_unit):
	time_series = generate_time_series(time_start, time_end, time_unit)
	if time_unit == 'daily':
		return [t.strftime('%H:%M') for t in time_series]
	elif time_unit == 'weekly':
		return [StringManager.get_week_name(t.weekday()) for t in time_series]
	elif time_unit == 'monthly':
		return [t.strftime('%-d') + (('\n' + StringManager.get_month_name(t.month)) if idx == 0 or t.day == 1 else '') for idx, t in enumerate(time_series)]
	else:
		return [StringManager.get_month_name(t.month) for t in time_series]

def estimate_zero_cross_value(time_list, free_list):
	if len(time_list) < 2:
		return -1
	elif free_list[-1] == free_list[0]:
		return -1

	return float(-free_list[-1]) / (free_list[-1] - free_list[0]) * (time_list[-1] - time_list[0])

def estimated_exhaust_remain(usage_list, data_from, data_to, time_unit):
	first_write_index = -1
	for i in range(len(usage_list)):
		if usage_list[i]['write_size'] > 0:
			first_write_index = max(0, i - 1)
			break

	time_value_list = [(t - data_from).total_seconds() for t in generate_time_series(data_from, data_to, time_unit)]
	valid_usage_list = usage_list[first_write_index:]
	total_write_size = sum([usage['write_size'] for usage in valid_usage_list])
	free_size_list = [usage['volume_free_space'] for usage in valid_usage_list]
	remain_seconds = estimate_zero_cross_value(time_value_list, free_size_list)

	if remain_seconds >= 0:
		return render_time_interval(data_to, remain_seconds)
	elif total_write_size == 0:
		return ('report:estimate_no_exauhst',)

	remain_seconds = float(free_size_list[-1]) / total_write_size * ((data_to - data_from).total_seconds())
	return render_time_interval(data_to, remain_seconds)

def render_time_interval(from_date, seconds):
	seconds_of_a_day = 24 * 60 * 60
	seconds_of_30days = 30 * seconds_of_a_day
	months = int(seconds / seconds_of_30days)
	remain_days = int((seconds - months * seconds_of_30days) / seconds_of_a_day)

	if 365 * seconds_of_a_day <= seconds:
		try:
			return (from_date + timedelta(seconds=seconds)).strftime('%Y-%m-%d')
		except OverflowError:
			return ('report:estimate_no_exauhst',)
	elif 1 < months:
		if 1 < remain_days:
			return ('report:months_and_days', months, remain_days)
		else:
			return ('report:months_and_day', months, remain_days)
	elif 1 == months:
		if 1 < remain_days:
			return ('report:month_and_days', months, remain_days)
		else:
			return ('report:month_and_day', months, remain_days)
	else:
		if 1 < remain_days:
			return ('common:days', remain_days)
		elif 1 == remain_days:
			return ('report:day', remain_days)
		else:
			return ('report:within_a_day', remain_days)

def render_device_name(value, record, headers):
	return value

def render_task_name(value, record, headers):
	if not value:
		return '--'
	else:
		return value

def getByNamespace(obj, namespace, default=None):
	value = reduce(lambda o, k: o.get(k, None) if isinstance(o, dict) else None, namespace.split('.'), obj)
	if value is None:
		return default
	return value

def render_device_name_in_param(value, record, headers):
	other_params = record[-1]
	device_name = other_params.get('device_name', None)
	if device_name is not None:
		return device_name
	device_list = getByNamespace(other_params, 'task_config.device_list', [])
	if device_list:
		return ', '.join([device['host_name'] for device in device_list])
	return '--'

def render_task_name_in_param(value, record, headers):
	return record[-1].get('task_name', '--')

def render_timestamp(value, record, headers):
	return datetime.fromtimestamp(value).strftime('%Y-%m-%d %H:%M:%S')

def render_last_backup_time(time_diff, record, headers, generate_time):
	if time_diff == generate_time:
		return ('report:nerver_backup_earlier',)
	else:
		if record[3] == 0: # last_backup_time
			return '--'

		days_ago = time_diff // 86400

		if days_ago < 1:
			return ('component:within_a_day',)

		return ('component:n_days_ago', days_ago)

def render_last_backup_time_data(data, headers, generate_time):
	if util.output_format == 'pdf':
		return [{
			'device_name': escape(render_device_name(row[0], row, headers)),
			'last_backup_task_name': render_task_name(row[1], row, headers),
			'time_diff': row[2],
			'last_backup_time': render_last_backup_time(row[2], row, headers, generate_time)
		} for row in data]
	else:
		return [{
			'device_name': render_device_name(row[0], row, headers),
			'last_backup_task_name': render_task_name(row[1], row, headers),
			'time_diff': row[2],
			'last_backup_time': render_last_backup_time(row[2], row, headers, generate_time)
		} for row in data]

def render_error_log_data(data, headers):
	if util.output_format == 'pdf':
		return [{
			'device_name': escape(render_device_name_in_param('', row, headers)),
			'task_name':   render_task_name_in_param('', row, headers),
			'log_time':    render_timestamp(row[0], row, headers),
			'log_type':    StringManager.get_log_string(row[1], row[2], False)
		} for row in data]
	else:
		return [{
			'device_name': render_device_name_in_param('', row, headers),
			'task_name':   render_task_name_in_param('', row, headers),
			'log_time':    render_timestamp(row[0], row, headers),
			'log_type':    (row[1], row[2])
		} for row in data]
