#!/usr/bin/env python
#-*-coding: utf-8 -*-
# Copyright (c) 2019 Synology Inc. All rights reserved.

from future import standard_library
standard_library.install_aliases()
standard_library.install_hooks()

import sys
from . import swinrm
import traceback
import os
import socket
import ssl
import errno
import requests

from http.client import BadStatusLine
from .utils import *
from .defines import *
from func_timeout import func_timeout, FunctionTimedOut
from .exceptions import (
    AuthenticateError,
    CerticateVerifiedError,
    PowerShellException
)

import logging
logger = logging.getLogger('synoabk-hypervisor-server')

class Main:
    def __init__(self, hostname, username, password, thumbprint, port, protocol, taskPool):
        self.hostname = hostname
        self.thumbprint = thumbprint
        self.auth = (username, password)
        self.protocol = protocol
        self.port = port

    def __del__(self):
        self.session = None
        logger.info('disconnect %s' % self.hostname)

    def dispatch(self, command, argument):
        command = getattr(self, command)
        argument = argument or {}
        response = {}

        always_trust_cert = False
        if 'always_trust_cert' in argument:
            always_trust_cert = argument['always_trust_cert']
            argument.pop('always_trust_cert')

        if 'force_connect' in argument:
            argument.pop('force_connect')

        if 'func_call_timeout' in argument:
            timeout = argument['func_call_timeout']
            argument.pop('func_call_timeout')
        else:
            timeout = 300

        try:
            func_timeout(300, self.connect, args=(always_trust_cert,))

            if 0 == timeout:
                response['data'] = command(**argument)
            else:
                response['data'] = func_timeout(timeout, command, kwargs=argument)
            response['success'] = True
        except BaseException as e:
            logger.error(traceback.format_exc())
            response = self._parse_exception(e)
            response['success'] = False

        return response

    def _parse_exception(self, e):
        error_class = type(e).__name__
        error_message = str(e.strerror) if hasattr(e, 'strerror') else str(e)
        error_params = {}

        for attr in dir(e):
            t = type(getattr(e, attr)).__name__
            if (t == 'str' or t == 'unicode' or t == 'int') and not attr.startswith("__"):
                error_params[attr] = getattr(e, attr)

        if type(e) == socket.error:
            if e.args[0] in (errno.EHOSTUNREACH, errno.ECONNREFUSED, errno.ETIMEDOUT, errno.EHOSTDOWN, errno.EPIPE, errno.ECONNRESET):
                error_class = 'ConnectionFail'
        elif type(e) in (BadStatusLine, ssl.SSLError, socket.timeout, requests.exceptions.SSLError, requests.ConnectionError, requests.Timeout):
            error_class = 'ConnectionFail'
        elif type(e) == FunctionTimedOut:
            error_message = 'WinRM operation timed out after %s seconds.' % ("N/A" if e.timedOutAfter == None else str(e.timedOutAfter))

        return {
            'error_class': error_class,
            'error_message': error_message,
            'error_params': error_params
        }

    def connect(self, always_trust_cert):
        if not self.auth[1]:
            raise AuthenticateError('Empty password')

        cert_validation = 'validate'
        ca_path = '/etc/ssl/certs'
        if True == always_trust_cert:
            cert_validation = 'ignore'
            ca_path = ''

        try:
            self.session = swinrm.Session(
                '%s://%s:%d' % (self.protocol, self.hostname, self.port),
                auth=self.auth,
                transport='ntlm',
                ca_trust_path=ca_path,
                server_cert_validation=cert_validation)

        except requests.exceptions.SSLError as e:
            cert_pem = ssl.get_server_certificate((self.hostname, self.port))
            current_thumbprint = self.getServerCert()['thumbprint']

            if current_thumbprint != self.thumbprint:
                reason = 'verify_failed'
                if "hostname" in str(e) and "doesn't match" in str(e):
                    reason = 'hostname_not_match'
                logger.error('certificate verified error!! except: %s, check_hostname: %s, on_disk_thumbprint: %s, current thumbprint: %s, always_trust_cert: %d, cert_validation: %s' % (str(e), self.hostname, self.thumbprint, current_thumbprint, always_trust_cert, cert_validation))
                raise CerticateVerifiedError(reason, certificate=cert_pem.encode())

            self.session = swinrm.Session(
                '%s://%s:%d' % (self.protocol, self.hostname, self.port),
                auth=self.auth,
                transport='ntlm',
                server_cert_validation='ignore')

        try:
            self.session.run_cmd('winrm set winrm/config/service \'@{MaxConcurrentOperationsPerUser="4294967295"}\'') # workaround
        except:
            logger.error(traceback.format_exc())
        logger.info('connect %s' % self.hostname)

    def isScriptCanRun(self):
        response = self.session.run_ps(
            'Try { $p="$env:temp/ActiveBackup-' + randomString() + '.ps1"; New-Item $p -ItemType "file" -Value " " | Out-Null; &$p } Catch { $_ } Finally { Remove-Item (Get-Item $p).FullName }'
        )
        response.raise_for_status()
        return not response.json

    def checkSCVMM(self):
        response = self.session.run_fps(
            'Get-SCVMMServer',
            clauses=[('ComputerName', 'localhost')],
            columns=['ProductVersion'])
        data = {}
        try:
            response.raise_for_status()
            data = response.json
            data['success'] = True
        except BaseException as e:
            data['success'] = False

        return data

    def getServerInfo(self, reconnect=False):
        columns = [
            'ProductID',
            'ProductVersion',
            ('ServerInterfaceVersion', ['Major', 'Minor'])
        ]

        response = self.session.run_fps(
            'Get-SCVMMServer',
            clauses=[('ComputerName', 'localhost')],
            columns=columns)
        #getServerInfo implement shouldn't fail
        #if the target machine failed to provide api's not provided out of box
        try:
            response.raise_for_status()
        except PowerShellException as e:
            pass
        data = response.json
        response = self.session.run_fps('(Get-WmiObject Win32_OperatingSystem).Caption')
        response.raise_for_status()
        data['host_version'] =response.json
        response = self.session.run_fps('(Get-WmiObject Win32_OperatingSystem).BuildNumber')
        response.raise_for_status()
        data['host_build_ver'] = response.json
        response = self.session.run_fps('Get-Wmiobject Win32_ComputerSystemProduct | Select-Object -ExpandProperty UUID')
        response.raise_for_status()
        data['instance_uuid'] = response.json

        return data

    def getServerCert(self):
        if self.protocol != 'https':
            return {'thumbprint': ''}
        import socket, ssl, hashlib
        context = ssl.SSLContext(ssl.PROTOCOL_SSLv23)

        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock = context.wrap_socket(sock, server_hostname=self.hostname)
        sock.connect((self.hostname, self.port))
        cert = sock.getpeercert(True)
        sock.close()

        sha1 = hashlib.sha1()
        sha1.update(cert)
        digest = sha1.hexdigest().lower()

        s, i = '', 0
        for c in digest:
            s += c
            i += 1
            if i % 2 == 0:
                s += ':'

        return {'thumbprint': s[:-1]}

    def refreshHostVM(self, host_id):
        cmd = 'Get-SCVMHost -ID "{}" | Read-SCVirtualMachine'.format(host_id)
        response = self.session.run_fps(
            cmd,
            json_depth=1,
            force_array=True)
        response.raise_for_status()
        return fixListFormat(response.json)

    def getHostVMId(self, host_id, vm_id):
        conditions = [
                ('VMId', '=', vm_id),
                ('VirtualizationPlatform', '=', 'HyperV')]
        cmd = 'Get-SCVMHost -ID "{}" | Get-SCVirtualMachine'.format(host_id)
        columns = ['ID','Name','VMId', 'Status']
        response = self.session.run_fps(
            cmd,
            conditions=conditions,
            columns=columns,
            json_depth=2,
            force_array=True)
        response.raise_for_status()
        return fixListFormat(response.json)

    def getVMs(self, parent_id = None, parent_type=None, vm_id=None, getDetail=False):
        cmd = 'Get-SCVirtualMachine'
        columns = [
            'Id',
            'VMId',
            'Name',
            'Memory',
            'CPUCount',
            'VirtualMachineState',
            'Status',
            'StatusString',
            'Generation',
            'TotalSize',
            'State',
            'Version',
            'HostId',
            'IsHighlyAvailable',
            ('OperatingSystem', ['Name', 'Architecture']),
            ('VMHost', ['FullyQualifiedDomainName', 'ID'])
        ]

        if getDetail:
            columns.extend([
                'VMCPath',
                ('VirtualHardDisks', ['Location', 'Size', 'ID', 'VHDType']),
                ('VirtualNetworkAdapters', ['VirtualNetwork', 'ID'])
            ])


        conditions = [('VirtualizationPlatform', '=', 'HyperV')]

        clauses=[]

        if parent_id:
            if parent_type == IdPrefixClusterNode:
                cmd = 'Get-SCVMHostCluster -ID "{}" | Get-SCVMHost | '.format(parent_id) + cmd
                conditions.append(('IsHighlyAvailable', '=', '$TRUE'))
            else:
                cmd = 'Get-SCVMHost -ID "{}" | '.format(parent_id) + cmd
                if parent_type == IdPrefixHostNode:
                    conditions.append(('IsHighlyAvailable', '=', '$FALSE'))

        if vm_id:
            clauses.append(('ID', vm_id))

        response = self.session.run_fps(
            cmd,
            clauses=clauses,
            conditions=conditions,
            columns=columns,
            json_depth=3,
            force_array=True)
        response.raise_for_status()

        out = fixListFormat(response.json)

        pathColumns = []

        if getDetail:
            pathColumns.extend([
                'VMCPath',
                ('VirtualHardDisks', ['Location'])
            ])

        dataToUnixPath(out, pathColumns)

        return out

    def resolveDnsName(self, fqdn):
        columns = [
            'IPAddress']
        conditions = [('DataLength', '=', '4')]
        result = self.session.run_fps(
            'resolve-dnsname -QuickTimeout "{}"'.format(fqdn),
            columns = columns,
            conditions = conditions,
            json_depth=1)
        result.raise_for_status()
        out = []
        for records in fixListFormat(result.json):
            out.append(records['IPAddress'])
        return out

    def getAllFolders(self):
        columns=[
                'ID',
                'Name',
                'IsRoot',
                ('ParentHostGroup', ['ID'])]

        result = self.session.run_fps(
            'Get-SCVMHostGroup',
            columns = columns,
            json_depth = 3,
            force_array=True)
        result.raise_for_status()
        folders = []
        data = fixListFormat(result.json)
        for folder in data:
            folder['type'] = 'vim.Folder'
            folder['parentId'] = '' if folder['IsRoot'] else folder['ParentHostGroup'][0]['ID']
            folders.append(folder)

        return folders

    def getOneHostInfo(self, hostId, hostName = '', fullyQualifiedDomainName = ''):
        hostNameList=[]
        if fullyQualifiedDomainName:
            fakehost = {}
            fakehost['FullyQualifiedDomainName'] = fullyQualifiedDomainName
            fakehost['ComputerName'] = hostName;
            hostNameList = [fakehost]
        else:
            conditions = [('VirtualizationPlatform', '=', 'HyperV')
                    ]

            columns=[
                    'ID',
                    'Name',
                    'ComputerName'
                    'FullyQualifiedDomainName']

            result = self.session.run_fps(
                'GET-SCVMHost',
                columns=columns,
                conditions=conditions,
                clauses=[('ID', hostId)],
                json_depth=3,
                force_array=True)
            result.raise_for_status()
            hostNameList = fixListFormat(result.json)

        hosts = []
        for host in hostNameList:
            ipList = []
            try:
                ipList = self.resolveDnsName(host['FullyQualifiedDomainName'])
            except BaseException as e:
                logger.error('host[%s] get ip failed[%s]' % (host['ComputerName'], str(e)))
            host['ip'] = None if len(ipList) == 0 else ipList[0]
            host['name'] = host['ComputerName'];
            hosts.append(host)
        return hosts
    def getAllHostSharePath(self):
        conditions = [('VirtualizationPlatform', '=', 'HyperV')
                ]

        columns=[
                'ID',
                'Name',
                'ComputerName',
                ('HostCluster',['SharedVolumes'])]

        result = self.session.run_fps(
            'GET-SCVMHost',
            columns=columns,
            conditions=conditions,
            json_depth=3,
            force_array=True)
        result.raise_for_status()

        hosts = {}
        data = fixListFormat(result.json)

        for host in data:
            try:
                if 'HostCluster' in host and 'SharedVolumes' in host['HostCluster'][0]:
                    sharePaths = host['HostCluster'][0]['SharedVolumes'].split(' ')
                    hosts[host['ID']] = [p.lower() for p in sharePaths]
            except BaseException as e:
                logger.error('host[%s] fix data failed[%s]' % (host['ComputerName'], str(host)))
        return hosts

    def getAllHosts(self, getIP):
        conditions = [('VirtualizationPlatform', '=', 'HyperV')
                ]

        columns=[
                'ID',
                'Name',
                'ComputerName',
                ('HostCluster',['HostGroup','ClusterName', 'ID']),
                ('VMHostGroup', ['Path', 'ID'])]

        if getIP:
            columns.append('FullyQualifiedDomainName')

        result = self.session.run_fps(
            'GET-SCVMHost',
            columns=columns,
            conditions=conditions,
            json_depth=3,
            force_array=True)
        result.raise_for_status()

        pathColumns = [
            ('HostCluster', ['HostGroup']),
            ('VMHostGroup', ['Path'])
        ]
        hosts = []
        data = fixListFormat(result.json)
        dataToUnixPath(data, pathColumns)
        for host in data:
            try:
                scvmm_path = host['ComputerName']
                parentId = ''
                if len(host['HostCluster']) > 0:
                    folderPath = host['HostCluster'][0]['HostGroup']
                    idx = folderPath.find('/')
                    if idx > -1:
                        folderPath = folderPath[idx:]
                    else:
                        folderPath = '/'
                    scvmm_path = os.path.join(
                            folderPath,
                            host['HostCluster'][0]['ClusterName'],
                            host['ComputerName'])
                    parentId = host['HostCluster'][0]['ID']
                elif len(host['VMHostGroup']) > 0:
                    folderPath = host['VMHostGroup'][0]['Path']
                    idx = folderPath.find('/')
                    if idx > -1:
                        folderPath = folderPath[idx:]
                    else:
                        folderPath = '/'
                    scvmm_path = os.path.join(
                            folderPath,
                            host['ComputerName'])
                    parentId = host['VMHostGroup'][0]['ID']
                host['scvmmPath'] = scvmm_path
                host['type'] = 'vim.ComputeResource'
                host['parentId'] = parentId
                if getIP:
                    ipList = []
                    try:
                        if host['FullyQualifiedDomainName']:
                            ipList = self.resolveDnsName(host['FullyQualifiedDomainName'])
                    except BaseException as e:
                        logger.error('host[%s] get ip failed[%s]' % (host['ComputerName'], str(e)))
                    host['ip'] = None if len(ipList) == 0 else ipList[0]
                hosts.append(host)
            except BaseException as e:
                logger.error('host[%s] fix data failed[%s]' % (host['ComputerName'], str(host)))
        return hosts

    def getAllClusters(self):
        conditions = [('VirtualizationPlatform', '=', 'HyperV')
                ]

        columns=[
                'ID',
                'Name',
                'ClusterName',
                ('HostGroup', ['Path', 'ID'])]
        result = self.session.run_fps(
            'Get-SCVMHostCluster',
            columns=columns,
            conditions=conditions,
            json_depth=3,
            force_array=True)
        result.raise_for_status()

        pathColumns = [
            ('HostGroup', ['Path']),
        ]
        clusters = []
        data = fixListFormat(result.json)
        dataToUnixPath(data, pathColumns)
        for cluster in data:
            try:
                scvmm_path = cluster['ClusterName']
                parentId = ''
                if len(cluster['HostGroup']) > 0:
                    folderPath = cluster['HostGroup'][0]['Path']
                    idx = folderPath.find('/')
                    if idx > -1:
                        folderPath = folderPath[idx:]
                    else:
                        folderPath = '/'
                    scvmm_path = os.path.join(
                            folderPath,
                            cluster['ClusterName'])
                    parentId = cluster['HostGroup'][0]['ID']
                cluster['scvmmPath'] = scvmm_path
                cluster['ComputerName'] = cluster['ClusterName']
                cluster['type'] = 'vim.ClusterComputeResource'
                cluster['parentId'] = parentId
                clusters.append(cluster)
            except BaseException as e:
                logger.error('cluster[%s] fix data failed[%s]' % (cluster['ClusterName'], str(cluster)))

        return clusters

    def listAllNodes(self, getCluster = False, getHost = False, getFolder = False, getIP = False):
        out = []

        if getHost:
            out += self.getAllHosts(getIP)
        if getCluster:
            out += self.getAllClusters()
        if getFolder:
            out += self.getAllFolders()

        return out

    def listNode(self, parent=None):
        def _rootGetNode():
            result = self.session.run_fps(
                'Get-SCVMHostGroup | Where IsRoot -eq True',
                columns=['ID', 'Name'])
            result.raise_for_status()
            return _folderGetNode(result.json['ID'])

        def _folderGetNode(parentId):
            conditions = [
                    ('VirtualizationPlatform', '=', 'HyperV'),
                    ('HostCluster', '=', '$NULL')]

            folders = self.session.run_fps(
                'Get-SCVMHostGroup -ID "{}" | Get-SCVMHostGroup'.format(parentId),
                columns=['Name','ID'],
                json_depth=2,
                force_array=True)
            folders.raise_for_status()

            clusters = self.session.run_fps(
                'Get-SCVMHostGroup -ID "{}" | Get-SCVMHostCluster'.format(parentId),
                columns=['ClusterName','ID'],
                conditions=conditions,
                json_depth=2,
                force_array=True)
            clusters.raise_for_status()

            hosts = self.session.run_fps(
                'Get-SCVMHostGroup -ID "{}" | Get-SCVMHost'.format(parentId),
                columns=['ComputerName','ID', 'FullyQualifiedDomainName'],
                conditions=conditions,
                json_depth=2,
                force_array=True)
            hosts.raise_for_status()

            data = {
                    'folders': fixListFormat(folders.json),
                    'clusters': fixListFormat(clusters.json),
                    'hosts': fixListFormat(hosts.json)
                    }
            return data

        def _clusterGetNode(parentId):
            conditions = [('VirtualizationPlatform', '=', 'HyperV')]
            response = self.session.run_fps(
                'Get-SCVMHostCluster -ID "{}" | Get-SCVMHost'.format(parentId),
                columns=['ComputerName','ID', 'FullyQualifiedDomainName'],
                conditions=conditions,
                json_depth=3,
                force_array=True)
            response.raise_for_status()

            data = {
                    'folders': [],
                    'clusters': [],
                    'hosts': fixListFormat(response.json)
                    }
            return data

        result = []
        if not parent or 'Type' not in parent or not parent['Type']:
            return _rootGetNode()
        elif parent['Type'] == IdPrefixFolderNode:
            return _folderGetNode(parent['ID'])
        elif parent['Type'] == IdPrefixClusterNode:
            return _clusterGetNode(parent['ID'])
        return result
    def unregisterVm(self, scvmm_id):
        cmd = 'Get-SCVirtualMachine -ID "{}" | Remove-SCVirtualMachine -SkipFileRemoval'.format(scvmm_id)
        result = self.session.run_fps(cmd, json_depth=1)
        result.raise_for_status()
        return len(fixListFormat(result.json))
    def removeVm(self, scvmm_id):
        cmd = 'Get-SCVirtualMachine -ID "{}" | Remove-SCVirtualMachine'.format(scvmm_id)
        result = self.session.run_fps(cmd, json_depth=1)
        result.raise_for_status()
        return len(fixListFormat(result.json))
    def poweronVm(self, scvmm_id):
        cmd = 'Get-SCVirtualMachine -ID "{}" | Start-SCVirtualMachine'.format(scvmm_id)
        result = self.session.run_fps(cmd, json_depth=1)
        result.raise_for_status()
        return len(fixListFormat(result.json))
    def poweroffVm(self, scvmm_id):
        cmd = 'Get-SCVirtualMachine -ID "{}" | Stop-SCVirtualMachine'.format(scvmm_id)
        result = self.session.run_fps(cmd, json_depth=1)
        result.raise_for_status()
        return len(fixListFormat(result.json))
