#-*-coding: utf-8 -*-
import os, fcntl
import uuid, json
import logging
import six
import logging.handlers
from .defines import *

from .log import (
    LogType,
    LogLevel,
    ActivityLog
)

from .exceptions import (
    GuestOSAttributeError
)

logger = logging.getLogger(VM_LOGGER_NAME)

def isSuccessStatusCode(code):
    return 200 <= code and code < 300

def toUnixPath(path):
    if not path:
        return ''
    if path[0] == '/':
        return path
    return path.replace('\\', '/')

def toWinPath(path):
    return path.replace('/', '\\')

def toUnixText(text):
    return text.replace('\r\n', '\n')

def escapePath(path):
    path = path.replace('`', '``')
    path = path.replace('$', '`$')
    path = path.replace('"', '""')
    return path

def toList(var):
    if var == None:
        return []
    if isinstance(var, list) == False:
        return [var]
    return var

def fixListFormat(var):
    if var == None:
        return []
    elif isinstance(var, dict) == True and len(var) == 0:
        return []
    elif isinstance(var, list) == True:
        return var
    return [var]

def dataToUnixPath(data, columns):
    for record in data:
        for key in columns:
            if type(key) is tuple:
                dataToUnixPath(record[key[0]], key[1])
            elif key in record:
                record[key] = toUnixPath(record[key])

def beUser(uid, gid):
    if os.getuid() != 0:
        return

    # Remove group privileges
    os.setgroups([])

    # Try setting the new uid/gid
    os.setgid(gid)
    os.setuid(uid)

    # Ensure a very conservative umask
    os.umask(0o77)

def predictRestoreMeta(isWindows, username):
    miss_attr_list = []

    miss_attr_list.append(ATTR_ATIME)
    miss_attr_list.append(ATTR_MTIME)
    if not isWindows:
        miss_attr_list.append(ATTR_PERMS)
        if username == "root":
            miss_attr_list.append(ATTR_UID)
            miss_attr_list.append(ATTR_GID)
    return miss_attr_list

def exponentialRetryCall(retry_max, exception_retry_condition):
    def decorator(func):
        def wrapper(*args, **kwargs):
            retry_count = 0;
            from time import sleep

            while True:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    if not exception_retry_condition or not exception_retry_condition(e):
                        import traceback
                        logger.error(traceback.format_exc())
                        raise e

                    logger.error('func: [%s] args: [%s] kwarg: [%s] exception: [%s][%s]' %
                            (six.get_function_code(func), str(args), str(kwargs), e.__class__.__name__, str(e)))

                    if (retry_count >= retry_max):
                        raise e

                    retry_count += 1
                    if retry_count < 9:
                        sleep_time = 10 + (2**retry_count)
                    else:
                        sleep_time = 10 + (2**8)

                    logger.error("func: [%s], retry: [%d] times, after [%d] secs" % \
                            (func.__name__, retry_count, sleep_time))
                    sleep(sleep_time)
        return wrapper
    return decorator

def getJsonFromString(string):
    json_start = string.find('{')
    json_end = string.rfind('}')
    if json_start == -1 or json_end == -1:
        raise ValueError('can not find valid json in string %s' % string)
    return json.loads(string[json_start:json_end+1])

class ProgressInfo:
    def __init__(self):
        self.state = 'running'
        self.error = None
        self.current_progress = 0
        self.current_percent = 0
        self.current_path = ''
        self.accumulated_progress = 0

class Progress:
    def __init__(self, progressFile, parse_exception, log_result_id):
        self.info = ProgressInfo()
        self._error = False
        self._successCount = 0
        self._isMetaChange = False
        self._progressFile = progressFile
        self._progressTempFile = progressFile + ".temp"
        self._parse_exception = parse_exception
        self._firstException = None
        self._updateCounter = 0
        self._updateProgress(True)
        self._log = ActivityLog(log_result_id)
        try:
            os.unlink(self._progressTempFile)
        except Exception as e:
            pass

    def startFile(self, file_path):
        self.info.accumulated_progress += self.info.current_progress
        self.info.current_progress = 0
        self.info.current_path = file_path
        self._updateProgress(True)
    def setProgress(self, progress, percent):
        self.info.current_progress = progress
        self.info.current_percent = percent
        self._updateProgress()
    def addDirSuccess(self, dir_path, is_update_meta):
        other_params = {}
        other_params["path"] = dir_path
        self._successCount += 1
        self._log.add_result_detail(LogType.PATH_RESTORE, other_params)
        if is_update_meta:
            self._isMetaChange = True
    def addFileSuccess(self, file_path, is_update_meta):
        other_params = {}
        other_params["path"] = file_path
        self._successCount += 1
        self._log.add_result_detail(LogType.PATH_RESTORE, other_params)
        if is_update_meta:
            self._isMetaChange = True
    def addFileSkipSuccess(self, file_path, file_size):
        self.setProgress(file_size, 100.0)
        self.addFileSuccess(file_path, False)
    def addErr(self, file_path, exception):
        #self._addErrReason(exception)  ==> refer to VmFailureReason()
        other_params = self._parse_exception(exception)
        other_params["path"] = file_path
        log_type = LogType.PATH_RESTORE_ERROR

        if type(exception) == GuestOSAttributeError:
            log_type = LogType.PATH_RESTORE_META_ERROR
            other_params["attrList"] = other_params["error_params"]["miss_attr_list"]
            self._log.add_result_detail(log_type, other_params)
        elif self._log.logLevel[log_type] == LogLevel.ERROR:
            if "error_params" in other_params:
                for key in other_params["error_params"]:
                    other_params[key] = other_params["error_params"][key]
            logType = self._getLogType(exception)
            if LogType.INTERNAL_ERROR != logType:
                self._log.add_result_detail(logType, other_params)
            self._log.add_result_detail(LogType.PATH_RESTORE_ERROR, other_params)
            self._error = True
            if not self._firstException:
                self._firstException = exception

    def _getLogType(self, exception):
        error_class = type(exception).__name__
        if error_class == "GuestNotEnoughSpace":
            return LogType.VM_RESTORE_GUEST_NOT_ENOUGH_SPACE
        elif error_class == "GuestOSException":
            return LogType.REASON_GUEST_OS
        elif error_class == "UnicodeEncodeError":
            return LogType.REASON_GUEST_OS
        elif error_class == "GuestOSAttributeError":
            return LogType.PATH_RESTORE_META_ERROR
        elif error_class == "PathRestoreNotSupport":
            return LogType.PATH_RESTORE_NOT_SUPPORT
        return LogType.INTERNAL_ERROR

    def _updateProgress(self, is_force = False):
        if not len(self._progressFile):
            return
        if not is_force and self._updateCounter < 100:
            self._updateCounter += 1
            return

        self._updateCounter = 0 #reset counter
        fdProgress = open(self._progressTempFile, "w")
        response = {}
        data = {}
        data["progress"] = self.info.accumulated_progress + self.info.current_progress
        data["state"] = self.info.state
        data["current_path"] = self.info.current_path
        data["success_count"] = self._successCount
        data["is_meta_change"] = self._isMetaChange

        if "success" == self.info.state or "running" == self.info.state:
            response["success"] = True
        elif "error" == self.info.state:
            response = self._parse_exception(self._firstException)
            response["success"] = False # Client should throw exception

        response["data"] = data
        fdProgress.write(json.dumps(response))
        fdProgress.close()
        os.rename(self._progressTempFile, self._progressFile)

    def end(self):
        if self._error:
            self.info.state = "error"
        else:
            self.info.state = "success"
        self._updateProgress(True)


class PrivilegeEnv:
    def __init__(self):
        pass
    def isCanceled(self):
        if "1" == os.getenv('IS_PRIVILEGE_WORKER_CANCELED'):
            return True
        else:
            return False
